from langchain.chains import LLMChain, SimpleSequentialChain
from langchain.prompts import PromptTemplate
from langchain.llms import OpenAI
import os
from log10.load import log10, log10_session
import openai

log10(openai)

openai.api_key = os.getenv("OPENAI_API_KEY")


llm = OpenAI(temperature=0.9, model_name="text-curie-001")

with log10_session():
    prompt = PromptTemplate(
        input_variables=["product"],
        template="What is a good name for a company that makes {product}?",
    )
    chain = LLMChain(llm=llm, prompt=prompt)
    second_prompt = PromptTemplate(
        input_variables=["company_name"],
        template="Write a catchphrase for the following company: {company_name}",
    )
    chain_two = LLMChain(llm=llm, prompt=second_prompt)
    overall_chain = SimpleSequentialChain(
        chains=[chain, chain_two], verbose=True)
    # Run the chain specifying only the input variable for the first chain.
    catchphrase = overall_chain.run("colorful socks")
    print(catchphrase)

with log10_session():
    third_prompt = PromptTemplate(
        input_variables=["month"],
        template="What is a good country to travel during {month}?",
    )

    chain_three = LLMChain(llm=llm, prompt=third_prompt)

    fourth_prompt = PromptTemplate(
        input_variables=["country_name"],
        template="Write a 1 day itinerary to {country_name}",
    )
    chain_four = LLMChain(llm=llm, prompt=fourth_prompt)

    overall_chain_two = SimpleSequentialChain(
        chains=[chain_three, chain_four], verbose=True)

    # Run the chain specifying only the input variable for the first chain.
    catchphrase_two = overall_chain_two.run("April")
    print(catchphrase_two)




import os
from log10.agents.scrape_summarizer import scrape_summarizer
from log10.load import log10


# Select one of OpenAI or Anthropic models
model = "gpt-3.5-turbo-16k"
# model = "claude-1"

if 'claude' in model:
    import anthropic
    log10(anthropic)
    anthropicClient = anthropic.Client(os.environ["ANTHROPIC_API_KEY"])
    module = anthropicClient
    hparams = {'max_tokens_to_sample': 1024}
    summary_model = "claude-1-100k"
else:  # openai
    import openai
    log10(openai)
    openai.api_key = os.getenv("OPENAI_API_KEY")
    hparams = {}
    module = openai
    summary_model = "gpt-3.5-turbo-16k"

url = "https://nytimes.com"
print(scrape_summarizer(url, model, module, hparams))



import faker
import sqlalchemy
from langchain.sql_database import SQLDatabase
from langchain.agents.agent_toolkits import SQLDatabaseToolkit
from langchain.agents import create_sql_agent
from langchain.llms import OpenAI
from faker import Faker
import random
import datetime
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine, Column, Integer, String, DateTime
import os
from log10.load import log10
import openai

log10(openai)


# Set up a dummy database
fake = Faker()

# Create a SQLite database and connect to it
engine = create_engine('sqlite:///users.db', echo=True)
Base = declarative_base()

# Define the User class with standard fields and the created_at field
class User(Base):
    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    username = Column(String, unique=True, nullable=False)
    email = Column(String, unique=True, nullable=False)
    first_name = Column(String, nullable=False)
    last_name = Column(String, nullable=False)
    age = Column(Integer, nullable=False)
    created_at = Column(DateTime, default=datetime.datetime.utcnow)

    def __repr__(self):
        return f"<User(id={self.id}, username='{self.username}', email='{self.email}', first_name='{self.first_name}', last_name='{self.last_name}', age={self.age}, created_at={self.created_at})>"

# Helper function to generate a random user using Faker
def generate_random_user():
    username = fake.user_name()
    email = fake.email()
    first_name = fake.first_name()
    last_name = fake.last_name()
    age = random.randint(18, 100)
    return User(username=username, email=email, first_name=first_name, last_name=last_name, age=age)


# Create the 'users' table
Base.metadata.create_all(engine)

# Create a session factory and a session
Session = sessionmaker(bind=engine)
session = Session()

# Add some example users
for n_users in range(10):
    user = generate_random_user()
    session.add(user)

session.commit()

# Query the users and print the results
all_users = session.query(User).all()
print(all_users)

session.close()

# Setup vars for Langchain
openai.api_key = os.getenv("OPENAI_API_KEY")

# Setup Langchain SQL agent
db = SQLDatabase.from_uri("sqlite:///users.db")
toolkit = SQLDatabaseToolkit(db=db)

agent_executor = create_sql_agent(
    llm=OpenAI(temperature=0, model_name="text-davinci-003"),
    toolkit=toolkit,
    verbose=True
)

print(agent_executor.run("Who is the least recent user?"))



import types
import functools
import inspect
import requests
import os
import json
import time
import traceback
from aiohttp import ClientSession
import asyncio
import threading
import queue
from contextlib import contextmanager
import logging
from dotenv import load_dotenv
import backoff  # for exponential backoff
from openai.error import RateLimitError, APIConnectionError

load_dotenv()

url = os.environ.get("LOG10_URL")
token = os.environ.get("LOG10_TOKEN")
org_id = os.environ.get("LOG10_ORG_ID")

# log10, bigquery
target_service = os.environ.get("LOG10_DATA_STORE")

if target_service == "bigquery":
    from log10.bigquery import initialize_bigquery
    bigquery_client, bigquery_table = initialize_bigquery()
    import uuid
    from datetime import datetime, timezone
elif target_service is None:
    target_service = "log10"  # default to log10


@backoff.on_exception(backoff.expo, (RateLimitError, APIConnectionError))
def func_with_backoff(func, *args, **kwargs):
    return func(*args, **kwargs)


def get_session_id():
    if target_service == "bigquery":
        return str(uuid.uuid4())

    try:
        session_url = url + "/api/sessions"
        res = requests.request("POST",
                               session_url, headers={"x-log10-token": token, "Content-Type": "application/json"}, json={
                                   "organization_id": org_id
                               })

        return res.json()['sessionID']
    except Exception as e:
        raise Exception("Failed to create LOG10 session: " + str(e) + "\nLikely cause: LOG10 env vars missing or not picked up correctly!" +
                        "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details")


# Global variable to store the current sessionID.
sessionID = get_session_id()
last_completion_response = None


class log10_session:
    def __enter__(self):
        global sessionID
        global last_completion_response
        sessionID = get_session_id()
        last_completion_response = None
        return self

    def last_completion_url(self):
        if last_completion_response is None:
            return None

        return url + '/app/' + last_completion_response['organizationSlug'] + '/completions/' + last_completion_response['completionID']

    def __exit__(self, exc_type, exc_value, traceback):
        return


@contextmanager
def timed_block(block_name):
    if DEBUG:
        start_time = time.perf_counter()
        try:
            yield
        finally:
            elapsed_time = time.perf_counter() - start_time
            logging.debug(
                f"TIMED BLOCK - {block_name} took {elapsed_time:.6f} seconds to execute.")
    else:
        yield


def log_url(res, completionID):
    output = res.json()
    organizationSlug = output['organizationSlug']
    full_url = url + '/app/' + organizationSlug + '/completions/' + completionID
    logging.debug(f"LOG10: Completion URL: {full_url}")


async def log_async(completion_url, func, **kwargs):
    async with ClientSession() as session:
        global last_completion_response

        res = requests.request("POST",
                               completion_url, headers={"x-log10-token": token, "Content-Type": "application/json"}, json={
                                   "organization_id": org_id
                               })
        # todo: handle session id for bigquery scenario
        last_completion_response = res.json()
        completionID = res.json()['completionID']

        if DEBUG:
            log_url(res, completionID)
        log_row = {
            # do we want to also store args?
            "status": "started",
            "orig_module": func.__module__,
            "orig_qualname": func.__qualname__,
            "request": json.dumps(kwargs),
            "session_id": sessionID,
            "organization_id": org_id
        }
        if target_service == "log10":
            res = requests.request("POST",
                                   completion_url + "/" + completionID,
                                   headers={"x-log10-token": token,
                                            "Content-Type": "application/json"},
                                   json=log_row)
        elif target_service == "bigquery":
            pass
            # NOTE: We only save on request finalization.

        return completionID


def run_async_in_thread(completion_url, func, result_queue, **kwargs):
    result = asyncio.run(
        log_async(completion_url=completion_url, func=func, **kwargs))
    result_queue.put(result)


def log_sync(completion_url, func, **kwargs):
    global last_completion_response
    res = requests.request("POST",
                           completion_url, headers={"x-log10-token": token, "Content-Type": "application/json"}, json={
                               "organization_id": org_id
                           })

    last_completion_response = res.json()
    completionID = res.json()['completionID']
    if DEBUG:
        log_url(res, completionID)
    res = requests.request("POST",
                           completion_url + "/" + completionID,
                           headers={"x-log10-token": token,
                                    "Content-Type": "application/json"},
                           json={
                               # do we want to also store args?
                               "status": "started",
                               "orig_module": func.__module__,
                               "orig_qualname": func.__qualname__,
                               "request": json.dumps(kwargs),
                               "session_id": sessionID,
                               "organization_id": org_id
                           })
    return completionID


def intercepting_decorator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        completion_url = url + "/api/completions"
        output = None
        result_queue = queue.Queue()

        try:
            with timed_block(sync_log_text + " call duration"):
                if USE_ASYNC:
                    threading.Thread(target=run_async_in_thread, kwargs={
                        "completion_url": completion_url, "func": func, "result_queue": result_queue, **kwargs}).start()
                else:
                    completionID = log_sync(
                        completion_url=completion_url, func=func, **kwargs)

            current_stack_frame = traceback.extract_stack()
            stacktrace = ([{"file": frame.filename,
                          "line": frame.line,
                           "lineno": frame.lineno,
                            "name": frame.name} for frame in current_stack_frame])

            start_time = time.perf_counter()
            output = func_with_backoff(func, *args, **kwargs)
            duration = time.perf_counter() - start_time
            logging.debug(
                f"LOG10: TIMED BLOCK - LLM call duration: {duration}")

            if USE_ASYNC:
                with timed_block("extra time spent waiting for log10 call"):
                    while result_queue.empty():
                        pass
                    completionID = result_queue.get()

            with timed_block("result call duration (sync)"):
                # Adjust the Anthropic output to match OAI completion output
                if func.__qualname__ == "Client.completion":
                    output['choices'] = [{
                        'text': output['completion'],
                        'index': 0,
                    }]
                    log_row = {
                        "response": json.dumps(output),
                        "status": "finished",
                        "duration": int(duration*1000),
                        "stacktrace": json.dumps(stacktrace),
                        "kind": "completion"
                    }
                else:
                    log_row = {
                        "response": json.dumps(output),
                        "status": "finished",
                        "duration": int(duration*1000),
                        "stacktrace": json.dumps(stacktrace)
                    }

                if target_service == "log10":
                    res = requests.request("POST",
                                           completion_url + "/" + completionID,
                                           headers={
                                               "x-log10-token": token, "Content-Type": "application/json"},
                                           json=log_row)
                elif target_service == "bigquery":
                    try:
                        log_row["id"] = str(uuid.uuid4())
                        log_row["created_at"] = datetime.now(
                            timezone.utc).isoformat()
                        log_row["request"] = json.dumps(kwargs)

                        if func.__qualname__ == "Completion.create":
                            log_row["kind"] = "completion"
                        elif func.__qualname__ == "ChatCompletion.create":
                            log_row["kind"] = "chat"

                        log_row["orig_module"] = func.__module__
                        log_row["orig_qualname"] = func.__qualname__
                        log_row["session_id"] = sessionID

                        bigquery_client.insert_rows_json(
                            bigquery_table, [log_row])

                    except Exception as e:
                        logging.error(
                            f"LOG10: failed to insert in Bigquery: {log_row} with error {e}")
        except Exception as e:
            logging.error("LOG10: failed", e)

        return output

    return wrapper


def set_sync_log_text(USE_ASYNC=True):
    return "async" if USE_ASYNC else "sync"


def log10(module, DEBUG_=False, USE_ASYNC_=True):
    """Intercept and overload module for logging purposes

    Keyword arguments:
    module -- the module to be intercepted (e.g. openai)
    DEBUG_ -- whether to show log10 related debug statements via python logging (default False)
    USE_ASYNC_ -- whether to run in async mode (default True)
    """
    global DEBUG, USE_ASYNC, sync_log_text
    DEBUG = DEBUG_
    USE_ASYNC = USE_ASYNC_
    sync_log_text = set_sync_log_text(USE_ASYNC=USE_ASYNC)
    logging.basicConfig(level=logging.DEBUG if DEBUG else logging.INFO,
                        format='%(asctime)s - %(levelname)s - LOG10 - %(message)s')

    # def intercept_nested_functions(obj):
    #     for name, attr in vars(obj).items():
    #         if callable(attr) and isinstance(attr, types.FunctionType):
    #             setattr(obj, name, intercepting_decorator(attr))
    #         elif inspect.isclass(attr):
    #             intercept_class_methods(attr)

    # def intercept_class_methods(cls):
    #     for method_name, method in vars(cls).items():
    #         if isinstance(method, classmethod):
    #             original_method = method.__func__
    #             decorated_method = intercepting_decorator(original_method)
    #             setattr(cls, method_name, classmethod(decorated_method))
    #         elif isinstance(method, (types.FunctionType, types.MethodType)):
    #             print(f"method:{method}")
    #             setattr(cls, method_name, intercepting_decorator(method))
    #         elif inspect.isclass(method):  # Handle nested classes
    #             intercept_class_methods(method)

    for name, attr in vars(module).items():
        if inspect.isclass(attr):
            # OpenAI
            if module.__name__ == "openai" and name in ["ChatCompletion", "Completion"]:
                for method_name, method in vars(attr).items():
                    if isinstance(method, classmethod):
                        original_method = method.__func__
                        if original_method.__qualname__ in ["ChatCompletion.create", "Completion.create"]:
                            decorated_method = intercepting_decorator(
                                original_method)
                            setattr(attr, method_name,
                                    classmethod(decorated_method))
            # Anthropic
            elif module.__name__ == "anthropic" and name == "Client":
                for method_name, method in vars(attr).items():
                    if isinstance(method, (types.FunctionType, types.MethodType)) and method_name == "completion":
                        setattr(attr, method_name,
                                intercepting_decorator(method))

            # For future reference:
            # if callable(attr) and isinstance(attr, types.FunctionType):
            #     print(f"attr:{attr}")
            #     setattr(module, name, intercepting_decorator(attr))
            # elif inspect.isclass(attr):  # Check if attribute is a class
            #     intercept_class_methods(attr)
            # # else: # uncomment if we want to include nested function support
            # #     intercept_nested_functions(attr)



from langchain.chains import LLMChain, SimpleSequentialChain
from langchain.prompts import PromptTemplate
from langchain.llms import OpenAI
import os
from log10.load import log10, log10_session
import openai

log10(openai)

openai.api_key = os.getenv("OPENAI_API_KEY")

llm = OpenAI(temperature=0.9, model_name="text-curie-001")

with log10_session() as session:
    print(session.last_completion_url())

    response = openai.Completion.create(
        model="text-ada-001",
        prompt="Why did the chicken cross the road?",
        temperature=0,
        max_tokens=1024,
        top_p=1,
        frequency_penalty=0,
        presence_penalty=0,
    )

    print(session.last_completion_url())

    response = openai.Completion.create(
        model="text-ada-001",
        prompt="Why did the cow cross the road?",
        temperature=0,
        max_tokens=1024,
        top_p=1,
        frequency_penalty=0,
        presence_penalty=0,
    )

    print(session.last_completion_url())

with log10_session() as session:
    print(session.last_completion_url())

    response = openai.Completion.create(
        model="text-ada-001",
        prompt="Why did the frog cross the road?",
        temperature=0,
        max_tokens=1024,
        top_p=1,
        frequency_penalty=0,
        presence_penalty=0,
    )

    print(session.last_completion_url())

    response = openai.Completion.create(
        model="text-ada-001",
        prompt="Why did the scorpion cross the road?",
        temperature=0,
        max_tokens=1024,
        top_p=1,
        frequency_penalty=0,
        presence_penalty=0,
    )

    print(session.last_completion_url())



import os
from log10.load import log10
from log10.evals import compile
from log10.agents.camel import camel_agent
from log10.tools import code_extractor

# Select one of OpenAI or Anthropic models
#model = "gpt-3.5-turbo-16k"
model = "claude-1"
maxTurns = 10

if 'claude' in model:
    import anthropic
    log10(anthropic)
    anthropicClient = anthropic.Client(os.environ["ANTHROPIC_API_KEY"])
    module = anthropicClient
    hparams = {'max_tokens_to_sample': 1024}
    summary_model = "claude-1-100k"
    extraction_model = "claude-1-100k"
    completion_func = anthropicClient.completion
else:  # openai
    import openai
    log10(openai)
    openai.api_key = os.getenv("OPENAI_API_KEY")
    hparams = {}
    module = openai
    summary_model = "gpt-3.5-turbo-16k"
    extraction_model = "gpt-4"
    completion_func = openai.ChatCompletion.create

# example calls from playground (select 1)
user_messages, assistant_messages = camel_agent(userRole='C developer', assistantRole='Cybersecurity expert',
                                                taskPrompt="Correct the following code.\n\n#include <stdio.h>\n#include <string.h>\n\nint main() {\n    char password[8];\n    int granted = 0;\n\n    printf(\"Enter password: \");\n    scanf(\"%s\", password);\n\n    if (strcmp(password, \"password\") == 0) {\n        granted = 1;\n    }\n\n    if (granted) {\n        printf(\"Access granted.\\n\");\n    } else {\n        printf(\"Access denied.\\n\");\n    }\n\n    return 0;\n}",
                                                model=model, summary_model=summary_model, maxTurns=maxTurns,
                                                module=module, hparams=hparams)

full_response = assistant_messages[-1]['content']

# Next extract just the C code
code = code_extractor(full_response, "C", completion_func, extraction_model, hparams)
print(f"Extracted code\n###\n{code}")

# Evaluate if the code compiles
result = compile(code)
if result is True:
    print("Compilation successful")
else:
    print("Compilation failed with error:")
    print(result[1])



import sys

if 'init_modules' in globals():
    # second or subsequent run: remove all but initially loaded modules
    for m in list(sys.modules.keys()):
        if m not in init_modules:
            del (sys.modules[m])
else:
    # first run: find out which modules were initially loaded
    init_modules = list(sys.modules.keys())

import os
from log10.load import log10, log10_session
import openai

openai.api_key = os.getenv("OPENAI_API_KEY")

# Launch an async run
with log10_session():
    log10(openai, DEBUG_=True, USE_ASYNC_=True)
    completion = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[
            {'role': "system", "content": "You are the most knowledgable Star Wars guru on the planet"},
            {"role": "user", "content": "Write the time period of all the Star Wars movies and spinoffs?"}
        ]
    )
    print(completion.choices[0].message)

# reload modules to prevent double calling openAI
if 'init_modules' in globals():
    # second or subsequent run: remove all but initially loaded modules
    for m in list(sys.modules.keys()):
        if m not in init_modules:
            del (sys.modules[m])
else:
    # first run: find out which modules were initially loaded
    init_modules = list(sys.modules.keys())


import openai  # noqa

# Compare to sync run - note there can be variability in the OpenAI calls
with log10_session():
    log10(openai, DEBUG_=True, USE_ASYNC_=False)

    completion = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[
            {'role': "system", "content": "You are the most knowledgable Star Wars guru on the planet"},
            {"role": "user", "content": "Write the time period of all the Star Wars movies and spinoffs?"}
        ]
    )
    print(completion.choices[0].message)



import os
from log10.load import log10
from log10.agents.camel import camel_agent
from dotenv import load_dotenv
load_dotenv()

# Select one of OpenAI or Anthropic models
model = "gpt-3.5-turbo-16k"
# model = "claude-1"
maxTurns = 30

if 'claude' in model:
    import anthropic
    log10(anthropic)
    anthropicClient = anthropic.Client(os.environ["ANTHROPIC_API_KEY"])
    module = anthropicClient
    hparams = {'max_tokens_to_sample': 1024}
    summary_model = "claude-1-100k"
else:  # openai
    import openai
    log10(openai)
    openai.api_key = os.getenv("OPENAI_API_KEY")
    hparams = {}
    module = openai
    summary_model = "gpt-3.5-turbo-16k"

# example calls from playground (select 1)
camel_agent(userRole='Sales email copyeditor', assistantRole='Sales email copywriter',
            taskPrompt='Write a sales email to Pfizer about a new healthcare CRM',
            model=model, summary_model=summary_model, maxTurns=maxTurns,
            module=module, hparams=hparams)



import os
from log10.load import log10
from log10.evals import eval

# Choose provider
provider = "anthropic" # "anthropic"

if provider == "openai":
    import openai
    log10(openai)
    openai.api_key = os.getenv("OPENAI_API_KEY")
    # Define completion spec
    completion_func = openai.ChatCompletion.create
    completion_params = {
        'model': "gpt-3.5-turbo",
        'temperature': 0,
        'max_tokens': 1024,
        'top_p': 1,
        'frequency_penalty': 0,
        'presence_penalty': 0,
        }
elif provider == "anthropic":
    import anthropic
    log10(anthropic)
    anthropicClient = anthropic.Client(os.environ["ANTHROPIC_API_KEY"])
    completion_func = anthropicClient.completion
    completion_params = {
        'model': 'claude-1',
        'temperature': 0,
        'max_tokens_to_sample': 1024,
        }
else:
    print(f"Unsupported provider option: {provider}. Supported providers are 'openai' or 'anthropic'.")

# Ground truth dataset to use for evaluation
eval_dataset = ('fuzzy_data.csv', {
                'input': 'my_input_column', 'ideal': 'my_output_column'})

# Specify which metrics to use. Options are:
# 'match': model_output.startswith(ideal)
# 'includes': ideal.lower() in model_output.lower()
# 'fuzzy_match': similar to includes but remove punctuation, articles and extra whitespace and compare both ways
eval_metric = 'fuzzy_match'

# Path to output file to store the metrics
# Example from: https://github.com/openai/evals/blob/a24f20a357ecb3cc5eec8323097aeade9585796c/evals/registry/evals/test-basic.yaml#L7
out_file_path = "fuzzy_output.csv"

# Get back and id and url for the summary of results and status
# todo: get back path to logfile; eval_id, eval_url =
eval(completion_func, completion_params, eval_dataset, eval_metric, out_file_path)



import os
from log10.load import log10
import openai

log10(openai)

openai.api_key = os.getenv("OPENAI_API_KEY")

completion = openai.ChatCompletion.create(
  model="gpt-3.5-turbo",
  messages=[
    {'role': "system", "content": "You are the most knowledgable Star Wars guru on the planet"},
    {"role": "user", "content": "Write the time period of all the Star Wars movies and spinoffs?"}
  ]
)

print(completion.choices[0].message)



import os
from collections import deque
from typing import Dict, List, Optional, Any

from langchain.chat_models import ChatOpenAI
from langchain import LLMChain, PromptTemplate
from langchain.embeddings import OpenAIEmbeddings
from langchain.llms import BaseLLM
from langchain.vectorstores.base import VectorStore
from pydantic import BaseModel, Field
from langchain.chains.base import Chain

from langchain.vectorstores import FAISS
from langchain.docstore import InMemoryDocstore

from langchain.agents import ZeroShotAgent, Tool, AgentExecutor
from langchain import OpenAI, SerpAPIWrapper, LLMChain

import openai
import log10
from log10.load import log10

log10(openai)

# Adapted from: https://python.langchain.com/en/latest/use_cases/agents/baby_agi_with_agent.html

# Define your embedding model
embeddings_model = OpenAIEmbeddings()
# Initialize the vectorstore as empty
import faiss
embedding_size = 1536
index = faiss.IndexFlatL2(embedding_size)
vectorstore = FAISS(embeddings_model.embed_query, index, InMemoryDocstore({}), {})


class TaskCreationChain(LLMChain):
    """Chain to generates tasks."""

    @classmethod
    def from_llm(cls, llm: BaseLLM, verbose: bool = True) -> LLMChain:
        """Get the response parser."""
        task_creation_template = (
            "You are an task creation AI that uses the result of an execution agent"
            " to create new tasks with the following objective: {objective},"
            " The last completed task has the result: {result}."
            " This result was based on this task description: {task_description}."
            " These are incomplete tasks: {incomplete_tasks}."
            " Based on the result, create new tasks to be completed"
            " by the AI system that do not overlap with incomplete tasks."
            " Return the tasks as an array."
        )
        prompt = PromptTemplate(
            template=task_creation_template,
            input_variables=["result", "task_description", "incomplete_tasks", "objective"],
        )
        return cls(prompt=prompt, llm=llm, verbose=verbose)
    
class TaskPrioritizationChain(LLMChain):
    """Chain to prioritize tasks."""

    @classmethod
    def from_llm(cls, llm: BaseLLM, verbose: bool = True) -> LLMChain:
        """Get the response parser."""
        task_prioritization_template = (
            "You are an task prioritization AI tasked with cleaning the formatting of and reprioritizing"
            " the following tasks: {task_names}."
            " Consider the ultimate objective of your team: {objective}."
            " Do not remove any tasks. Return the result as a numbered list, like:"
            " #. First task"
            " #. Second task"
            " Start the task list with number {next_task_id}."
        )
        prompt = PromptTemplate(
            template=task_prioritization_template,
            input_variables=["task_names", "next_task_id", "objective"],
        )
        return cls(prompt=prompt, llm=llm, verbose=verbose)

todo_prompt = PromptTemplate.from_template("You are a planner who is an expert at coming up with a todo list for a given objective. Come up with a todo list for this objective: {objective}")
todo_chain = LLMChain(llm=ChatOpenAI(temperature=0), prompt=todo_prompt)
search = SerpAPIWrapper()
tools = [
    Tool(
        name = "Search",
        func=search.run,
        description="useful for when you need to answer questions about current events"
    ),
    Tool(
        name = "TODO",
        func=todo_chain.run,
        description="useful for when you need to come up with todo lists. Input: an objective to create a todo list for. Output: a todo list for that objective. Please be very clear what the objective is!"
    )
]


prefix = """You are an AI who performs one task based on the following objective: {objective}. Take into account these previously completed tasks: {context}."""
suffix = """Question: {task}
{agent_scratchpad}"""
prompt = ZeroShotAgent.create_prompt(
    tools, 
    prefix=prefix, 
    suffix=suffix, 
    input_variables=["objective", "task", "context","agent_scratchpad"]
)
    
def get_next_task(task_creation_chain: LLMChain, result: Dict, task_description: str, task_list: List[str], objective: str) -> List[Dict]:
    """Get the next task."""
    incomplete_tasks = ", ".join(task_list)
    response = task_creation_chain.run(result=result, task_description=task_description, incomplete_tasks=incomplete_tasks, objective=objective)
    new_tasks = response.split('\n')
    return [{"task_name": task_name} for task_name in new_tasks if task_name.strip()]

def prioritize_tasks(task_prioritization_chain: LLMChain, this_task_id: int, task_list: List[Dict], objective: str) -> List[Dict]:
    """Prioritize tasks."""
    task_names = [t["task_name"] for t in task_list]
    next_task_id = int(this_task_id) + 1
    response = task_prioritization_chain.run(task_names=task_names, next_task_id=next_task_id, objective=objective)
    new_tasks = response.split('\n')
    prioritized_task_list = []
    for task_string in new_tasks:
        if not task_string.strip():
            continue
        task_parts = task_string.strip().split(".", 1)
        if len(task_parts) == 2:
            task_id = task_parts[0].strip()
            task_name = task_parts[1].strip()
            prioritized_task_list.append({"task_id": task_id, "task_name": task_name})
    return prioritized_task_list

def _get_top_tasks(vectorstore, query: str, k: int) -> List[str]:
    """Get the top k tasks based on the query."""
    results = vectorstore.similarity_search_with_score(query, k=k)
    if not results:
        return []
    sorted_results, _ = zip(*sorted(results, key=lambda x: x[1], reverse=True))
    return [str(item.metadata['task']) for item in sorted_results]

def execute_task(vectorstore, execution_chain: LLMChain, objective: str, task: str, k: int = 5) -> str:
    """Execute a task."""
    context = _get_top_tasks(vectorstore, query=objective, k=k)
    return execution_chain.run(objective=objective, context=context, task=task)

class BabyAGI(Chain, BaseModel):
    """Controller model for the BabyAGI agent."""

    task_list: deque = Field(default_factory=deque)
    task_creation_chain: TaskCreationChain = Field(...)
    task_prioritization_chain: TaskPrioritizationChain = Field(...)
    execution_chain: AgentExecutor = Field(...)
    task_id_counter: int = Field(1)
    vectorstore: VectorStore = Field(init=False)
    max_iterations: Optional[int] = None
        
    class Config:
        """Configuration for this pydantic object."""
        arbitrary_types_allowed = True

    def add_task(self, task: Dict):
        self.task_list.append(task)

    def print_task_list(self):
        print("\033[95m\033[1m" + "\n*****TASK LIST*****\n" + "\033[0m\033[0m")
        for t in self.task_list:
            print(str(t["task_id"]) + ": " + t["task_name"])

    def print_next_task(self, task: Dict):
        print("\033[92m\033[1m" + "\n*****NEXT TASK*****\n" + "\033[0m\033[0m")
        print(str(task["task_id"]) + ": " + task["task_name"])

    def print_task_result(self, result: str):
        print("\033[93m\033[1m" + "\n*****TASK RESULT*****\n" + "\033[0m\033[0m")
        print(result)
        
    @property
    def input_keys(self) -> List[str]:
        return ["objective"]
    
    @property
    def output_keys(self) -> List[str]:
        return []

    def _call(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Run the agent."""
        objective = inputs['objective']
        first_task = inputs.get("first_task", "Make a todo list")
        self.add_task({"task_id": 1, "task_name": first_task})
        num_iters = 0
        while True:
            if self.task_list:
                self.print_task_list()

                # Step 1: Pull the first task
                task = self.task_list.popleft()
                self.print_next_task(task)

                # Step 2: Execute the task
                result = execute_task(
                    self.vectorstore, self.execution_chain, objective, task["task_name"]
                )
                this_task_id = int(task["task_id"])
                self.print_task_result(result)

                # Step 3: Store the result in Pinecone
                result_id = f"result_{task['task_id']}"
                self.vectorstore.add_texts(
                    texts=[result],
                    metadatas=[{"task": task["task_name"]}],
                    ids=[result_id],
                )

                # Step 4: Create new tasks and reprioritize task list
                new_tasks = get_next_task(
                    self.task_creation_chain, result, task["task_name"], [t["task_name"] for t in self.task_list], objective
                )
                for new_task in new_tasks:
                    self.task_id_counter += 1
                    new_task.update({"task_id": self.task_id_counter})
                    self.add_task(new_task)
                self.task_list = deque(
                    prioritize_tasks(
                        self.task_prioritization_chain, this_task_id, list(self.task_list), objective
                    )
                )
            num_iters += 1
            if self.max_iterations is not None and num_iters == self.max_iterations:
                print("\033[91m\033[1m" + "\n*****TASK ENDING*****\n" + "\033[0m\033[0m")
                break
        return {}

    @classmethod
    def from_llm(
        cls,
        llm: BaseLLM,
        vectorstore: VectorStore,
        verbose: bool = False,
        **kwargs
    ) -> "BabyAGI":
        """Initialize the BabyAGI Controller."""
        task_creation_chain = TaskCreationChain.from_llm(
            llm, verbose=verbose
        )
        task_prioritization_chain = TaskPrioritizationChain.from_llm(
            llm, verbose=verbose
        )
        llm_chain = LLMChain(llm=llm, prompt=prompt)
        tool_names = [tool.name for tool in tools]
        agent = ZeroShotAgent(llm_chain=llm_chain, allowed_tools=tool_names)
        agent_executor = AgentExecutor.from_agent_and_tools(agent=agent, tools=tools, verbose=True)
        return cls(
            task_creation_chain=task_creation_chain,
            task_prioritization_chain=task_prioritization_chain,
            execution_chain=agent_executor,
            vectorstore=vectorstore,
            **kwargs
        )

OBJECTIVE = "Invent a new drug to cure cancer."

llm = ChatOpenAI(temperature=0)

# Logging of LLMChains
verbose=False
# If None, will keep on going forever
max_iterations: Optional[int] = 3
baby_agi = BabyAGI.from_llm(
    llm=llm,
    vectorstore=vectorstore,
    verbose=verbose,
    max_iterations=max_iterations
)

baby_agi({"objective": OBJECTIVE})




from log10.utils import convert_history_to_claude
from anthropic import HUMAN_PROMPT
from log10.tools import browser


# Set up Summarizer agent
system_prompt = "You are an expert at extracting the main points from a website. Only look at the provided website content by the user to extract the main points."
summarize_prompt = "Extract the main points from the following website:\n {website_text}"


def scrape_summarizer(url, model, module, hparams):
    website_text = browser(url)
    prompt = summarize_prompt.format(website_text=website_text)
    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": prompt}
    ]

    if 'claude' in model:
        completion = module.completion(
            prompt=convert_history_to_claude(messages),
            stop_sequences=[HUMAN_PROMPT],
            model=model,
            **hparams
        )
        website_summary = completion['completion']
    else:
        completion = module.ChatCompletion.create(
            model=model, messages=messages, temperature=0.2)
        website_summary = completion.choices[0].message
    return website_summary



import subprocess
import tempfile
import os
import csv
import json
import logging
from log10.utils import fuzzy_match, convert_history_to_claude, parse_field

def run_completion(example, completion_func, completion_params):
    funcname = completion_func.__qualname__
    try:
        if funcname == 'ChatCompletion.create':
            response = completion_func(**completion_params, messages=json.loads(example['input']))
            model_completion = response['choices'][0]['message']['content']
        elif funcname == 'Completion.create':
            response = completion_func(**completion_params, prompt=json.loads(example['input']))
            model_completion = response['choices'][0]['text']
        elif funcname == 'Client.completion':
            prompt = convert_history_to_claude(json.loads(example['input']))
            response = completion_func(**completion_params, prompt=prompt)
            model_completion = response['completion']
        else:
            logging.error(f"Unknown completion type {funcname}. Supported types: OpenAI (ChatCompletion.create, Completion.create) and Anthropic (Client.completion).")
        return model_completion
    except Exception as e:
        logging.error("LOG10 eval: failed", e)
    

def run_metric(metric, ideal, model_output):
    ideal = parse_field(ideal)
    for ideal_candidate in ideal:
        # Ref: https://github.com/openai/evals/blob/a24f20a357ecb3cc5eec8323097aeade9585796c/evals/elsuite/utils.py
        if metric == 'match':
            if model_output.startswith(ideal_candidate):
                return True
        elif metric == 'includes':  # case-insensitive
            # Ref: https://github.com/openai/evals/blob/a24f20a357ecb3cc5eec8323097aeade9585796c/evals/elsuite/basic/includes.py
            if ideal_candidate.lower() in model_output.lower():
                return True
        elif metric == 'fuzzy_match':
            if fuzzy_match(ideal_candidate, model_output):
                return True 
    return False


def write_to_csv(file_name, row_data):
    with open(file_name, 'a', newline='') as file:
        writer = csv.writer(file)
        writer.writerow(row_data)


def eval(completion_func, completion_params, dataset, metric, out_file_path):
    csv_file_name, mapping = dataset
    with open(csv_file_name, 'r') as csv_file:
        reader = csv.DictReader(csv_file)
        examples = []
        for example in reader:
            mapped_example = {}
            for key, value in mapping.items():
                mapped_example[key] = example[value]
            examples.append(mapped_example)

        # todo: each example could be launched as separate threads or separate api calls to job runners
        write_to_csv(out_file_path, [
                     'input', 'ideal', 'model_completion', 'metric'])
        for example in examples:
            model_completion = run_completion(
                example, completion_func, completion_params)
            example_metric = run_metric(
                metric, example['ideal'], model_completion)
            write_to_csv(out_file_path, [
                         example['input'], example['ideal'], model_completion, example_metric])
            print(
                f"\n\nIdeal:{example['ideal']}\nModel output:{model_completion}")
            print(f"Correct:{example_metric}")
    return

def compile(code):
    with tempfile.NamedTemporaryFile(suffix=".c", delete=False) as temp:
        temp.write(code.encode())
        temp.close()
        process = subprocess.run(["gcc", temp.name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        os.remove(temp.name)  # remove the temp file
        if process.returncode == 0:
            return True
        else:
            return False, process.stderr.decode()



import os
from log10.load import log10
import openai

log10(openai)

openai.api_key = os.getenv("OPENAI_API_KEY")

from langchain.llms import OpenAI
from langchain.prompts import PromptTemplate
from langchain.chains import LLMChain, SimpleSequentialChain

llm = OpenAI(temperature=0.9, model_name="text-babbage-001")
prompt = PromptTemplate(
    input_variables=["product"],
    template="What is a good name for a company that makes {product}?",
)

chain = LLMChain(llm=llm, prompt=prompt)

second_prompt = PromptTemplate(
    input_variables=["company_name"],
    template="Write a catchphrase for the following company: {company_name}",
)
chain_two = LLMChain(llm=llm, prompt=second_prompt)

overall_chain = SimpleSequentialChain(chains=[chain, chain_two], verbose=True)

# Run the chain specifying only the input variable for the first chain.
catchphrase = overall_chain.run("colorful socks")
print(catchphrase)


import os
from log10.load import log10
import openai

log10(openai)

openai.api_key = os.getenv("OPENAI_API_KEY")

response = openai.Completion.create(
  model="text-davinci-003",
  prompt="Write the names of all Star Wars movies and spinoffs along with the time periods in which they were set?",
  temperature=0,
  max_tokens=1024,
  top_p=1,
  frequency_penalty=0,
  presence_penalty=0
)

print(response)



import os
from log10.load import log10
import openai

log10(openai)

openai.api_key = os.getenv("OPENAI_API_KEY")
MAX_TOKENS = 512
TOOLS_DEFAULT_LIST =  ['llm-math', 'wikipedia']

from langchain.llms import OpenAI
from langchain.agents import load_tools, initialize_agent
import wikipedia

llm = OpenAI(temperature=0, model_name="text-davinci-003", max_tokens=MAX_TOKENS)

# Set up Langchain
tools = load_tools(TOOLS_DEFAULT_LIST, llm=llm)
chain = initialize_agent(tools, llm, agent="zero-shot-react-description", verbose=True)

inp = "How many years elapsed between the founding of Apple and Google?"
print(chain.run(inp))


import os
from log10.load import log10
import openai

log10(openai, DEBUG_=True, USE_ASYNC_=False)

openai.api_key = os.getenv("OPENAI_API_KEY")

response = openai.Completion.create(
    model="text-ada-001",
    prompt="What is 2+2?",
    temperature=0,
    max_tokens=1024,
    top_p=1,
    frequency_penalty=0,
    presence_penalty=0
)

print(response)



import os
from log10.load import log10
import openai

log10(openai)

openai.api_key = os.getenv("OPENAI_API_KEY")

# Example from: https://python.langchain.com/en/latest/use_cases/question_answering.html
# Download the state_of_the_union.txt here: https://raw.githubusercontent.com/hwchase17/langchain/master/docs/modules/state_of_the_union.txt
# This example requires: pip install chromadb

# Load Your Documents
from langchain.document_loaders import TextLoader
loader = TextLoader('../state_of_the_union.txt')

# Create Your Index
from langchain.indexes import VectorstoreIndexCreator
from langchain.vectorstores import Chroma
from langchain.embeddings import OpenAIEmbeddings
from langchain.text_splitter import CharacterTextSplitter
from langchain.chat_models import ChatOpenAI

index = VectorstoreIndexCreator(
    vectorstore_cls=Chroma, 
    embedding=OpenAIEmbeddings(),
    text_splitter=CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
).from_loaders([loader])

# Query Your Index
query = "What did the president say about Ketanji Brown Jackson"
print(index.query_with_sources(query, llm=ChatOpenAI(temperature=0, model_name="gpt-3.5-turbo")))



#!/usr/bin/env python

from distutils.core import setup
from setuptools import setup, find_packages

setup(
    name="Log10",
    version="0.0.9",
    description="Log10 LLM data management",
    author="Log10 team",
    author_email="team@log10.io",
    url="",
    packages=find_packages(),
    install_requires=[
        "openai",
        "python-dotenv",
    ],
    extras_require={
        "bigquery": ["google-cloud-bigquery"],
    },
)



from langchain.document_loaders import WebBaseLoader
from log10.utils import convert_history_to_claude

# Browser agent / tool
def browser(URL: str) -> str:
    """useful when you need to scrape a website. 
    Input is the URL of the website to be scraped.
    Output is the text of the website
    """
    # Example code if we want to follow links
    # explored_sites = set()
    # queue = [URL]

    # def parse_site(base_url):
    #   if base_url in explored_sites:
    #     return

    #   explored_sites.add(base_url)

    #   if debug:
    #     print("Scraping: ", base_url)

    #   page = requests.get(URL)
    #   soup = BeautifulSoup(page.content, "html.parser")
    #   for a_href in soup.find_all("a", href=True):
    #       link = a_href["href"]
    #       if link[0] == "/":
    #         link = URL + link

    #       if link.startswith(URL):
    #         to_explore = link.split("#")[0]
    #         queue.append(to_explore)

    # while len(queue) != 0:
    #   url = queue.pop()
    #   parse_site(url)

    # return list(explored_sites)

    loader = WebBaseLoader(URL)
    data = loader.load()
    data[0].page_content = " ".join(data[0].page_content.split())

    # hack: crop to 5000 char to fit within context length
    # Will need better strategy eventually
    #   maxlen = 5000 if len(data[0].page_content) > 5000 else len(data[0].page_content)
    #   return data[0].page_content[:maxlen]
    return data[0].page_content


def code_extractor(full_response, language, completion_func, extraction_model, hparams):
    """useful when you need to extract just the code from a detailed LLM response
    """
    funcname = completion_func.__qualname__
    messages = [
            {"role": "system", "content": f"Extract just the {language} code from the following snippet. Do not include ``` or markdown syntax. Format your reply so I can directly copy your entire response, save it as a file and compile and run the code."},
            {"role": "user", "content": "Here's the snippet:\n" + full_response}
        ]
    if funcname == 'ChatCompletion.create':
        completion = completion_func(model=extraction_model, messages=messages, temperature=0.2, **hparams)
        code = completion.choices[0].message.content
    elif funcname == 'Client.completion':
        prompt = convert_history_to_claude(messages)
        completion = completion_func(model=extraction_model, prompt=prompt, temperature=0.2, **hparams)
        code = completion['completion']

    return code



import os
from log10.load import log10
import anthropic
import os

log10(anthropic, DEBUG_=False)
anthropicClient = anthropic.Client(os.environ["ANTHROPIC_API_KEY"])

response = anthropicClient.completion(
    model="claude-1",
    prompt=f"\n\nHuman:Write the names of all Star Wars movies and spinoffs along with the time periods in which they were set?{anthropic.AI_PROMPT}",
    temperature=0,
    max_tokens_to_sample=1024,
    top_p=1,
    top_k=0
)

print(response)




import os
from log10.load import log10
from log10.evals import compile
from log10.tools import code_extractor
from log10.utils import convert_history_to_claude

# Select one of OpenAI or Anthropic models
model = "gpt-3.5-turbo"
#model = "claude-1"

if 'claude' in model:
    import anthropic
    log10(anthropic)
    anthropicClient = anthropic.Client(os.environ["ANTHROPIC_API_KEY"])
    module = anthropicClient
    hparams = {'max_tokens_to_sample': 1024}
    extraction_model = "claude-1-100k"
    completion_func = anthropicClient.completion
else:  # openai
    import openai
    log10(openai)
    openai.api_key = os.getenv("OPENAI_API_KEY")
    hparams = {}
    module = openai
    extraction_model = "gpt-4"
    completion_func = openai.ChatCompletion.create

# First, write a hello world program
messages = [
    {"role": "system", "content": "You are an expert C programmer."},
    {"role": "user", "content": "Write a hello world program. Insert a null character after the hello world"}
]

if 'claude' in model:
    prompt = convert_history_to_claude(messages)
    completion = anthropicClient.completion(prompt=prompt, model=model, **hparams)
    full_response = completion['completion']
else:
    completion = openai.ChatCompletion.create(model=model, messages=messages, temperature=0.2, **hparams)
    full_response = completion.choices[0].message.content
print(f"Full response\n###\n{full_response}")

# Next extract just the C code
code = code_extractor(full_response, "C", completion_func, extraction_model, hparams)
print(f"Extracted code\n###\n{code}")

# Evaluate if the code compiles
result = compile(code)
if result is True:
    print("Compilation successful")
else:
    print("Compilation failed with error:")
    print(result[1])
